/*  Copyright (C) 2011-2012  Povilas Kanapickas <povilas@radix.lt>

    Distributed under the Boost Software License, Version 1.0.
        (See accompanying file LICENSE_1_0.txt or copy at
            http://www.boost.org/LICENSE_1_0.txt)
*/

#include "../utils/test_helpers.h"
#include "../utils/test_results.h"
#include <simdpp/simd.h>
#include <limits>

namespace SIMDPP_ARCH_NAMESPACE {

#define TEST_COMPARE_TESTER_HELPER(TC, T, L, R)                     \
{                                                                   \
    TEST_ARRAY_HELPER2(TC, T, cmp_eq, L, R);                        \
    TEST_ARRAY_HELPER2(TC, T, cmp_lt, L, R);                        \
    TEST_ARRAY_HELPER2(TC, T, cmp_gt, L, R);                        \
}

template<unsigned B>
void test_compare_n(TestSuite& tc)
{
    using namespace simdpp;

    using uint8_n = uint8<B>;
    using int8_n = int8<B>;
    using uint16_n = uint16<B/2>;
    using int16_n = int16<B/2>;
    using uint32_n = uint32<B/4>;
    using int32_n = int32<B/4>;
    using uint64_n = uint64<B/8>;
    using int64_n = int64<B/8>;
    using float32_n = float32<B/4>;
    using float64_n = float64<B/8>;

    //int8x32
    {
        TestData<uint8_n> sl(
            make_uint(0x11, 0x22, 0x33, 0xff),
            make_uint(0xcc, 0xdd, 0xee, 0xff),
            make_uint(0x01, 0x00, 0xff, 0xfe)
        );
        TestData<uint8_n> sr(
            make_uint(0x22, 0x22, 0x22, 0x22),
            make_uint(0xee, 0xee, 0xee, 0xee),
            make_uint(0x00, 0x00, 0x00, 0x00)
        );

        TEST_COMPARE_TESTER_HELPER(tc, int8_n, sl, sr);
        TEST_COMPARE_TESTER_HELPER(tc, uint8_n, sl, sr);
    }

    //int16_n
    {
        TestData<uint16_n> sl(
            make_uint(0x1111, 0x2222, 0x3333, 0x3333),
            make_uint(0xdddd, 0xeeee, 0xffff, 0xcccc),
            make_uint(0x0001, 0x0000, 0xffff, 0xfffe)
        );
        TestData<uint16_n> sr(
            make_uint(0x2222, 0x2222, 0x2222, 0x2222),
            make_uint(0xeeee, 0xeeee, 0xeeee, 0xeeee),
            make_uint(0x0000, 0x0000, 0x0000, 0x0000)
        );

        TEST_COMPARE_TESTER_HELPER(tc, int16_n, sl, sr);
        TEST_COMPARE_TESTER_HELPER(tc, uint16_n, sl, sr);
    }

    //int32_n
    {
        TestData<uint32_n> sl(
            make_uint(0x11111111, 0x22222222, 0x33333333, 0x33333333),
            make_uint(0xdddddddd, 0xeeeeeeee, 0xffffffff, 0xcccccccc),
            make_uint(0x00000000, 0x00000001, 0xffffffff, 0xfffffffe)
        );

        TestData<uint32_n> sr(
            make_uint(0x22222222, 0x22222222, 0x22222222, 0x22222222),
            make_uint(0xeeeeeeee, 0xeeeeeeee, 0xeeeeeeee, 0xeeeeeeee),
            make_uint(0x00000000, 0x00000000, 0x00000000, 0x00000000)
        );

        TEST_COMPARE_TESTER_HELPER(tc, int32_n, sl, sr);
        TEST_COMPARE_TESTER_HELPER(tc, uint32_n, sl, sr);
    }

    tc.sync_archs();
#if SIMDPP_USE_NULL || SIMDPP_USE_AVX2 || SIMDPP_USE_NEON64
    //int64_n
    {
        TestData<uint64_n> sl(
            make_uint(0x1111111111111111, 0x2222222222222222, 0x3333333333333333, 0x3333333333333333),
            make_uint(0xdddddddddddddddd, 0xeeeeeeeeeeeeeeee, 0xffffffffffffffff, 0xcccccccccccccccc),
            make_uint(0x0000000000000000, 0x0000000000000001, 0xffffffffffffffff, 0xfffffffffffffffe)
        );
        TestData<uint64_n> sr(
            make_uint(0x2222222222222222),
            make_uint(0xeeeeeeeeeeeeeeee),
            make_uint(0x0000000000000000)
        );

        TEST_COMPARE_TESTER_HELPER(tc, int64_n, sl, sr);
        TEST_COMPARE_TESTER_HELPER(tc, uint64_n, sl, sr);
    }
#endif
    tc.sync_archs();

    float nanf = std::numeric_limits<float>::quiet_NaN();
    double nan = std::numeric_limits<double>::quiet_NaN();
    float inff = std::numeric_limits<float>::infinity();
    double inf = std::numeric_limits<double>::infinity();

    //float32_n
    {
        TestData<float32_n> sl(
            make_float(0.0f, 1.0f, 2.0f, inff),
            make_float(0.0f, 1.0f, 2.0f, inff),
            make_float(-0.0f, -1.0f, -2.0f, -inff),
            make_float(0.0f, 1.0f, -inff, inff),
            make_float(0.0f, 1.0f, -0.0f, -1.0f),
            make_float(nanf, 0.0, -inff, inff)
        );

        TestData<float32_n> sr(
            make_float(1.0f, 1.0f, 1.0f, 1.0f),
            make_float(-1.0f, -1.0f, -1.0f, -1.0f),
            make_float(-1.0f, -1.0f, -1.0f, -1.0f),
            make_float(inff, inff, inff, inff),
            make_float(nanf, nanf, nanf, nanf),
            make_float(nanf, nanf, nanf, nanf)
        );

        TEST_COMPARE_TESTER_HELPER(tc, float32_n, sl, sr);
        TEST_ARRAY_HELPER2(tc, float32_n, cmp_neq, sl, sr);
        TEST_ARRAY_HELPER2(tc, float32_n, cmp_le, sl, sr);
        TEST_ARRAY_HELPER2(tc, float32_n, cmp_ge, sl, sr);
    }

    //float64_n
    {
        TestData<float64_n> sl(
            make_float(0.0, 1.0),
            make_float(2.0, inf),
            make_float(0.0, 1.0),
            make_float(2.0, inf),
            make_float(-0.0, -1.0),

            make_float(-2.0, -inf),
            make_float(0.0, 1.0),
            make_float(-inf, inf),
            make_float(0.0, 1.0),
            make_float(-0.0, -1.0),

            make_float(nan, 0.0),
            make_float(-inf, inf)
        );

        TestData<float64_n> sr(
            make_float(1.0, 1.0),
            make_float(.0, 1.0),
            make_float(-1.0, -1.0),
            make_float(-1.0, -1.0),
            make_float(-1.0, -1.0),

            make_float(-1.0, -1.0),
            make_float(inf, inf),
            make_float(inf, inf),
            make_float(nan, nan),
            make_float(nan, nan),

            make_float(nan, nan),
            make_float(nan, nan)
        );

        TEST_COMPARE_TESTER_HELPER(tc, float64_n, sl, sr);
        TEST_ARRAY_HELPER2(tc, float64_n, cmp_neq, sl, sr);
        TEST_ARRAY_HELPER2(tc, float64_n, cmp_le, sl, sr);
        TEST_ARRAY_HELPER2(tc, float64_n, cmp_ge, sl, sr);
    }
}

void test_compare(TestResults& res)
{
    TestSuite& tc = NEW_TEST_SUITE(res, "compare");
    test_compare_n<16>(tc);
    test_compare_n<32>(tc);
    test_compare_n<64>(tc);
}

} // namespace SIMDPP_ARCH_NAMESPACE
